﻿/**
 * Copyright(c) Live2D Inc. All rights reserved.
 *
 * Use of this source code is governed by the Live2D Open Software license
 * that can be found at https://www.live2d.com/eula/live2d-open-software-license-agreement_en.html.
 */

#include "CubismOffscreenSurface_GXM.hpp"
#include <kernel.h>

namespace {
    Live2D::Cubism::Framework::csmBool s_areShadersReady = false;
    SceGxmShaderPatcher *s_shaderPatcher;
}

//------------ LIVE2D NAMESPACE ------------
namespace Live2D { namespace Cubism { namespace Framework { namespace Rendering {

CubismOffscreenFrame_GXM::CubismOffscreenFrame_GXM()
    : _offscreenColorBufferData(NULL)
    , _offscreenColorSurface(NULL)
    , _offscreenTexture(NULL)
    , _offscreenDepthBufferData(NULL)
    , _offscreenDepthSurface(NULL)
    , _offscreenRenderTarget(NULL)
    , _bufferWidth(0)
    , _bufferHeight(0)
    , _clearDataStorage(NULL)
    , _clearColorParam(NULL)
    , _clearVertexProgram(NULL)
    , _clearFragmentProgram(NULL)
    , _clearVertexProgramId(0)
    , _clearFragmentProgramId(0)
{
}

CubismOffscreenFrame_GXM::~CubismOffscreenFrame_GXM()
{
    if (_clearDataStorage != NULL) {
        sceGxmFreeDeviceMemLinux(_clearDataStorage);

        sceGxmShaderPatcherReleaseVertexProgram(s_shaderPatcher, _clearVertexProgram);
        sceGxmShaderPatcherReleaseFragmentProgram(s_shaderPatcher, _clearFragmentProgram);

        sceGxmShaderPatcherUnregisterProgram(s_shaderPatcher, _clearVertexProgramId);
        sceGxmShaderPatcherUnregisterProgram(s_shaderPatcher, _clearFragmentProgramId);

        s_shaderPatcher = NULL;
        s_areShadersReady = false;
    }
}

void CubismOffscreenFrame_GXM::BeginDraw(SceGxmContext **immContext)
{
    if (_offscreenRenderTarget == 0)
    {
        return;
    }

    // set up a scene, offscreen render target, no sync required
    sceGxmBeginScene(
        *immContext,
        0,
        _offscreenRenderTarget,
        NULL,
        NULL,
        NULL,
        _offscreenColorSurface,
        NULL);
}

void CubismOffscreenFrame_GXM::EndDraw(SceGxmContext **immContext)
{
    if (_offscreenRenderTarget == NULL)
    {
        return;
    }

    // stop rendering to the offscreen render target
    sceGxmEndScene(*immContext, NULL, NULL);
}

void CubismOffscreenFrame_GXM::Clear(float r, float g, float b, float a, SceGxmContext **immContext)
{
    // set clear shaders
    sceGxmSetVertexProgram(*immContext, _clearVertexProgram);
    sceGxmSetFragmentProgram(*immContext, _clearFragmentProgram);

    // set the fragment program constants
    void *fragmentDefaultBuffer;
    sceGxmReserveFragmentDefaultUniformBuffer(*immContext, &fragmentDefaultBuffer);
    float clearColor[4] = { r, g, b, a };
    sceGxmSetUniformDataF(fragmentDefaultBuffer, _clearColorParam, 0, 4, clearColor);

    // draw geometry
    sceGxmSetVertexStream(*immContext, 0, _clearDataStorage->mappedBase);
    sceGxmDraw(*immContext, SCE_GXM_PRIMITIVE_TRIANGLES, SCE_GXM_INDEX_FORMAT_U16, (unsigned char *)_clearDataStorage->mappedBase + 24, 3);
}

csmBool CubismOffscreenFrame_GXM::CreateOffscreenFrame(csmUint32 displayBufferWidth, csmUint32 displayBufferHeight, SceGxmShaderPatcher **shaderPatcher, SceGxmMultisampleMode *msaa)
{
    if (!s_areShadersReady) {
        // register programs with the shader patcher
        sceGxmShaderPatcherRegisterProgram(*shaderPatcher, VertShaderClearSrc, &_clearVertexProgramId);
        sceGxmShaderPatcherRegisterProgram(*shaderPatcher, FragShaderClearSrc, &_clearFragmentProgramId);

        // find attributes by name to create vertex format bindings
        const SceGxmProgramParameter *paramPositionAttribute = sceGxmProgramFindParameterByName(VertShaderClearSrc, "aPosition");

        // find fragment uniforms by name and cache parameter info
        _clearColorParam = sceGxmProgramFindParameterByName(FragShaderClearSrc, "color");

        // create clear vertex format
        SceGxmVertexAttribute clearVertexAttributes[1];
        SceGxmVertexStream clearVertexStreams[1];
        clearVertexAttributes[0].streamIndex = 0;
        clearVertexAttributes[0].offset = 0;
        clearVertexAttributes[0].format = SCE_GXM_ATTRIBUTE_FORMAT_F32;
        clearVertexAttributes[0].componentCount = 2;
        clearVertexAttributes[0].regIndex = sceGxmProgramParameterGetResourceIndex(paramPositionAttribute);
        clearVertexStreams[0].stride = 8;
        clearVertexStreams[0].indexSource = SCE_GXM_INDEX_SOURCE_INDEX_16BIT;

        // create clear programs
        sceGxmShaderPatcherCreateVertexProgram(
            *shaderPatcher,
            _clearVertexProgramId,
            clearVertexAttributes,
            1,
            clearVertexStreams,
            1,
            &_clearVertexProgram);

        sceGxmShaderPatcherCreateFragmentProgram(
            *shaderPatcher,
            _clearFragmentProgramId,
            SCE_GXM_OUTPUT_REGISTER_FORMAT_UCHAR4,
            *msaa,
            NULL,
            VertShaderClearSrc,
            &_clearFragmentProgram);

        // allocate vertices and indices

        sceGxmAllocDeviceMemLinux(
            SCE_GXM_DEVICE_HEAP_ID_USER_NC,
            SCE_GXM_MEMORY_ATTRIB_READ,
            4 * 1024,
            4096,
            &_clearDataStorage);

        float *commonStorage = (float *)_clearDataStorage->mappedBase;

        // write vertex data
        commonStorage[0] = -1.0f;//x
        commonStorage[1] = -1.0f;//y
        commonStorage[2] = 3.0f;//x
        commonStorage[3] = -1.0f;//y
        commonStorage[4] = -1.0f;//x
        commonStorage[5] = 3.0f;//y

        uint16_t *indices = (uint16_t *)&commonStorage[6];

        // write index data
        indices[0] = 0;
        indices[1] = 1;
        indices[2] = 2;

        s_shaderPatcher = *shaderPatcher;
        s_areShadersReady = true;
    }

    // 一旦削除
    DestroyOffscreenFrame();

    // allocate memory
    sceGxmAllocDeviceMemLinux(
        SCE_GXM_DEVICE_HEAP_ID_CDRAM,
        SCE_GXM_MEMORY_ATTRIB_READ | SCE_GXM_MEMORY_ATTRIB_WRITE,
        displayBufferWidth * displayBufferHeight * 4,
        MAX(SCE_GXM_TEXTURE_ALIGNMENT, SCE_GXM_COLOR_SURFACE_ALIGNMENT),
        &_offscreenColorBufferData);

    _offscreenColorSurface = new SceGxmColorSurface;
    // set up the surface
    sceGxmColorSurfaceInit(
        _offscreenColorSurface,
        SCE_GXM_COLOR_FORMAT_A8B8G8R8,
        SCE_GXM_COLOR_SURFACE_LINEAR,
        (*msaa == SCE_GXM_MULTISAMPLE_NONE) ? SCE_GXM_COLOR_SURFACE_SCALE_NONE : SCE_GXM_COLOR_SURFACE_SCALE_MSAA_DOWNSCALE,
        SCE_GXM_OUTPUT_REGISTER_SIZE_32BIT,
        displayBufferWidth,
        displayBufferHeight,
        displayBufferWidth,
        _offscreenColorBufferData->mappedBase);

    _offscreenTexture = new SceGxmTexture;
    // set up the texture
    sceGxmTextureInitLinear(
        _offscreenTexture,
        _offscreenColorBufferData->mappedBase,
        SCE_GXM_TEXTURE_FORMAT_A8B8G8R8,
        displayBufferWidth,
        displayBufferHeight,
        1);

    // set linear filtering
    sceGxmTextureSetMagFilter(_offscreenTexture, SCE_GXM_TEXTURE_FILTER_LINEAR);
    sceGxmTextureSetMinFilter(_offscreenTexture, SCE_GXM_TEXTURE_FILTER_LINEAR);
    
    // set up parameters
    SceGxmRenderTargetParams params;
    memset(&params, 0, sizeof(SceGxmRenderTargetParams));
    params.flags = 0;
    params.width = displayBufferWidth;
    params.height = displayBufferHeight;
    params.scenesPerFrame = 1;
    params.multisampleMode = *msaa;
    params.multisampleLocations = 0;
    params.driverMemBlock = SCE_UID_INVALID_UID;

    // create the render target
    sceGxmCreateRenderTarget(&params, &_offscreenRenderTarget);

    _bufferWidth = displayBufferWidth;
    _bufferHeight = displayBufferHeight;

    // 成功
    return true;
}

void CubismOffscreenFrame_GXM::DestroyOffscreenFrame()
{
    if (_offscreenRenderTarget != NULL)
    { 
        sceGxmDestroyRenderTarget(_offscreenRenderTarget);
        sceGxmFreeDeviceMemLinux(_offscreenColorBufferData);
        sceGxmFreeDeviceMemLinux(_offscreenDepthBufferData);

        delete _offscreenColorSurface;
        delete _offscreenTexture;
        delete _offscreenDepthSurface;

        _offscreenRenderTarget = NULL;
    }
}

SceGxmTexture *CubismOffscreenFrame_GXM::GetColorBuffer()
{
    return _offscreenTexture;
}

csmUint32 CubismOffscreenFrame_GXM::GetBufferWidth() const
{
    return _bufferWidth;
}

csmUint32 CubismOffscreenFrame_GXM::GetBufferHeight() const
{
    return _bufferHeight;
}

csmBool CubismOffscreenFrame_GXM::IsValid() const
{
    return _offscreenRenderTarget != NULL;
}

}}}}

//------------ LIVE2D NAMESPACE ------------
