﻿/**
 * Copyright(c) Live2D Inc. All rights reserved.
 *
 * Use of this source code is governed by the Live2D Open Software license
 * that can be found at https://www.live2d.com/eula/live2d-open-software-license-agreement_en.html.
 */

#include "CubismRenderer_GXM.hpp"
#include "Math/CubismMatrix44.hpp"
#include "Type/csmVector.hpp"
#include "Model/CubismModel.hpp"
#include <float.h>
#include <kernel.h>

//------------ LIVE2D NAMESPACE ------------
namespace Live2D { namespace Cubism { namespace Framework { namespace Rendering {

/*********************************************************************************************************************
*                                      CubismClippingManager_GXM
********************************************************************************************************************/
///< ファイルスコープの変数宣言
namespace {
const csmInt32 ColorChannelCount = 4;   ///< 実験時に1チャンネルの場合は1、RGBだけの場合は3、アルファも含める場合は4
}

CubismClippingManager_GXM::CubismClippingManager_GXM() :
                                                                   _currentFrameNo(0)
                                                                   , _clippingMaskBufferSize(256)
{
    CubismRenderer::CubismTextureColor* tmp;
    tmp = CSM_NEW CubismRenderer::CubismTextureColor();
    tmp->R = 1.0f;
    tmp->G = 0.0f;
    tmp->B = 0.0f;
    tmp->A = 0.0f;
    _channelColors.PushBack(tmp);
    tmp = CSM_NEW CubismRenderer::CubismTextureColor();
    tmp->R = 0.0f;
    tmp->G = 1.0f;
    tmp->B = 0.0f;
    tmp->A = 0.0f;
    _channelColors.PushBack(tmp);
    tmp = CSM_NEW CubismRenderer::CubismTextureColor();
    tmp->R = 0.0f;
    tmp->G = 0.0f;
    tmp->B = 1.0f;
    tmp->A = 0.0f;
    _channelColors.PushBack(tmp);
    tmp = CSM_NEW CubismRenderer::CubismTextureColor();
    tmp->R = 0.0f;
    tmp->G = 0.0f;
    tmp->B = 0.0f;
    tmp->A = 1.0f;
    _channelColors.PushBack(tmp);

}

CubismClippingManager_GXM::~CubismClippingManager_GXM()
{
    for (csmUint32 i = 0; i < _clippingContextListForMask.GetSize(); i++)
    {
        if (_clippingContextListForMask[i]) CSM_DELETE_SELF(CubismClippingContext, _clippingContextListForMask[i]);
        _clippingContextListForMask[i] = NULL;
    }

    // _clippingContextListForDrawは_clippingContextListForMaskにあるインスタンスを指している。上記の処理により要素ごとのDELETEは不要。
    for (csmUint32 i = 0; i < _clippingContextListForDraw.GetSize(); i++)
    {
        _clippingContextListForDraw[i] = NULL;
    }

    for (csmUint32 i = 0; i < _channelColors.GetSize(); i++)
    {
        if (_channelColors[i]) CSM_DELETE(_channelColors[i]);
        _channelColors[i] = NULL;
    }
}

void CubismClippingManager_GXM::Initialize(CubismModel& model, csmInt32 drawableCount, const csmInt32** drawableMasks, const csmInt32* drawableMaskCounts)
{
    //クリッピングマスクを使う描画オブジェクトを全て登録する
    //クリッピングマスクは、通常数個程度に限定して使うものとする
    for (csmInt32 i = 0; i < drawableCount; i++)
    {
        if (drawableMaskCounts[i] <= 0)
        {
            //クリッピングマスクが使用されていないアートメッシュ（多くの場合使用しない）
            _clippingContextListForDraw.PushBack(NULL);
            continue;
        }

        // 既にあるClipContextと同じかチェックする
        CubismClippingContext* cc = FindSameClip(drawableMasks[i], drawableMaskCounts[i]);
        if (cc == NULL)
        {
            // 同一のマスクが存在していない場合は生成する
            cc = CSM_NEW CubismClippingContext(this, drawableMasks[i], drawableMaskCounts[i]);
            _clippingContextListForMask.PushBack(cc);
        }

        cc->AddClippedDrawable(i);

        _clippingContextListForDraw.PushBack(cc);
    }
}

CubismClippingContext* CubismClippingManager_GXM::FindSameClip(const csmInt32* drawableMasks, csmInt32 drawableMaskCounts) const
{
    // 作成済みClippingContextと一致するか確認
    for (csmUint32 i = 0; i < _clippingContextListForMask.GetSize(); i++)
    {
        CubismClippingContext* cc = _clippingContextListForMask[i];
        const csmInt32 count = cc->_clippingIdCount;
        if (count != drawableMaskCounts) continue; //個数が違う場合は別物
        csmInt32 samecount = 0;

        // 同じIDを持つか確認。配列の数が同じなので、一致した個数が同じなら同じ物を持つとする。
        for (csmInt32 j = 0; j < count; j++)
        {
            const csmInt32 clipId = cc->_clippingIdList[j];
            for (csmInt32 k = 0; k < count; k++)
            {
                if (drawableMasks[k] == clipId)
                {
                    samecount++;
                    break;
                }
            }
        }
        if (samecount == count)
        {
            return cc;
        }
    }
    return NULL; //見つからなかった
}

void CubismClippingManager_GXM::SetupClippingContext(CubismModel& model, CubismRenderer_GXM* renderer)
{
    _currentFrameNo++;

    // 全てのクリッピングを用意する
    // 同じクリップ（複数の場合はまとめて１つのクリップ）を使う場合は１度だけ設定する
    csmInt32 usingClipCount = 0;
    for (csmUint32 clipIndex = 0; clipIndex < _clippingContextListForMask.GetSize(); clipIndex++)
    {
        // １つのクリッピングマスクに関して
        CubismClippingContext* cc = _clippingContextListForMask[clipIndex];

        // このクリップを利用する描画オブジェクト群全体を囲む矩形を計算
        CalcClippedDrawTotalBounds(model, cc);

        if (cc->_isUsing)
        {
            usingClipCount++; //使用中としてカウント
        }
    }

    // マスク作成処理
    if (usingClipCount > 0)
    {
        if (!renderer->IsUsingHighPrecisionMask())
        {
            csmFloat32 xScale = _clippingMaskBufferSize >> 1;
            csmFloat32 xOffset = xScale;
            csmFloat32 yScale = -(_clippingMaskBufferSize >> 1);
            csmFloat32 yOffset = _clippingMaskBufferSize + yScale;

            // モデル描画時にDrawMeshNowに渡される変換（モデルtoワールド座標変換）
            CubismMatrix44 modelToWorldF = renderer->GetMvpMatrix();

            // _offscreenFrameBufferへ切り替え
            renderer->_offscreenFrameBuffer.BeginDraw(&renderer->immContext_GXM);

            // マスクをクリアする
            // 1が無効（描かれない）領域、0が有効（描かれる）領域。（シェーダで Cd*Csで0に近い値をかけてマスクを作る。1をかけると何も起こらない）
            renderer->_offscreenFrameBuffer.Clear(1.0f, 1.0f, 1.0f, 1.0f, &renderer->immContext_GXM);
            sceGxmSetViewport(renderer->immContext_GXM, xOffset, xScale, yOffset, -yScale, 0.5f, 0.5f);
        }

        // 各マスクのレイアウトを決定していく
        SetupLayoutBounds(renderer->IsUsingHighPrecisionMask() ? 0 : usingClipCount);

        // 実際にマスクを生成する
        // 全てのマスクをどの様にレイアウトして描くかを決定し、ClipContext , ClippedDrawContext に記憶する
        for (csmUint32 clipIndex = 0; clipIndex < _clippingContextListForMask.GetSize(); clipIndex++)
        {
            // --- 実際に１つのマスクを描く ---
            CubismClippingContext* clipContext = _clippingContextListForMask[clipIndex];
            csmRectF* allClippedDrawRect = clipContext->_allClippedDrawRect; //このマスクを使う、全ての描画オブジェクトの論理座標上の囲み矩形
            csmRectF* layoutBoundsOnTex01 = clipContext->_layoutBounds; //この中にマスクを収める

            // モデル座標上の矩形を、適宜マージンを付けて使う
            const csmFloat32 MARGIN = 0.05f;
            _tmpBoundsOnModel.SetRect(allClippedDrawRect);
            _tmpBoundsOnModel.Expand(allClippedDrawRect->Width * MARGIN, allClippedDrawRect->Height * MARGIN);
            //########## 本来は割り当てられた領域の全体を使わず必要最低限のサイズがよい

            // シェーダ用の計算式を求める。回転を考慮しない場合は以下のとおり
            // movePeriod' = movePeriod * scaleX + offX     [[ movePeriod' = (movePeriod - tmpBoundsOnModel.movePeriod)*scale + layoutBoundsOnTex01.movePeriod ]]
            const csmFloat32 scaleX = layoutBoundsOnTex01->Width / _tmpBoundsOnModel.Width;
            const csmFloat32 scaleY = layoutBoundsOnTex01->Height / _tmpBoundsOnModel.Height;

            // マスク生成時に使う行列を求める
            {
                // シェーダに渡す行列を求める <<<<<<<<<<<<<<<<<<<<<<<< 要最適化（逆順に計算すればシンプルにできる）
                _tmpMatrix.LoadIdentity();
                {
                    // Layout0..1 を -1..1に変換
                    _tmpMatrix.TranslateRelative(-1.0f, -1.0f);
                    _tmpMatrix.ScaleRelative(2.0f, 2.0f);
                }
                {
                    // view to Layout0..1
                    _tmpMatrix.TranslateRelative(layoutBoundsOnTex01->X, layoutBoundsOnTex01->Y); //new = [translate]
                    _tmpMatrix.ScaleRelative(scaleX, scaleY); //new = [translate][scale]
                    _tmpMatrix.TranslateRelative(-_tmpBoundsOnModel.X, -_tmpBoundsOnModel.Y);
                    //new = [translate][scale][translate]
                }
                // tmpMatrixForMask が計算結果
                _tmpMatrixForMask.SetMatrix(_tmpMatrix.GetArray());
            }

            //--------- draw時の mask 参照用行列を計算
            {
                // シェーダに渡す行列を求める <<<<<<<<<<<<<<<<<<<<<<<< 要最適化（逆順に計算すればシンプルにできる）
                _tmpMatrix.LoadIdentity();
                {
                    _tmpMatrix.TranslateRelative(layoutBoundsOnTex01->X, layoutBoundsOnTex01->Y); //new = [translate]
                    _tmpMatrix.ScaleRelative(scaleX, scaleY); //new = [translate][scale]
                    _tmpMatrix.TranslateRelative(-_tmpBoundsOnModel.X, -_tmpBoundsOnModel.Y);
                    //new = [translate][scale][translate]
                }

                _tmpMatrixForDraw.SetMatrix(_tmpMatrix.GetArray());
            }

            clipContext->_matrixForMask.SetMatrix(_tmpMatrixForMask.GetArray());

            clipContext->_matrixForDraw.SetMatrix(_tmpMatrixForDraw.GetArray());

            if (!renderer->IsUsingHighPrecisionMask())
            {
                const csmInt32 clipDrawCount = clipContext->_clippingIdCount;
                for (csmInt32 i = 0; i < clipDrawCount; i++)
                {
                    const csmInt32 clipDrawIndex = clipContext->_clippingIdList[i];

                    // 頂点情報が更新されておらず、信頼性がない場合は描画をパスする
                    if (!model.GetDrawableDynamicFlagVertexPositionsDidChange(clipDrawIndex))
                    {
                        continue;
                    }

                    renderer->IsCulling(model.GetDrawableCulling(clipDrawIndex) != 0);

                    // 今回専用の変換を適用して描く
                    // チャンネルも切り替える必要がある(A,R,G,B)
                    renderer->SetClippingContextBufferForMask(clipContext);
                    renderer->DrawMesh(
                        model.GetDrawableTextureIndices(clipDrawIndex),
                        model.GetDrawableVertexIndexCount(clipDrawIndex),
                        model.GetDrawableVertexCount(clipDrawIndex),
                        const_cast<csmUint16*>(model.GetDrawableVertexIndices(clipDrawIndex)),
                        const_cast<csmFloat32*>(model.GetDrawableVertices(clipDrawIndex)),
                        reinterpret_cast<csmFloat32*>(const_cast<Core::csmVector2*>(model.GetDrawableVertexUvs(clipDrawIndex))),
                        model.GetDrawableOpacity(clipDrawIndex),
                        CubismRenderer::CubismBlendMode_Normal,   //クリッピングは通常描画を強制
                        false   // マスク生成時はクリッピングの反転使用は全く関係がない
                    );
                }
            }
        }

        if (!renderer->IsUsingHighPrecisionMask())
        {
            // --- 後処理 ---
            renderer->_offscreenFrameBuffer.EndDraw(&renderer->immContext_GXM); // 描画対象を戻す
            renderer->SetClippingContextBufferForMask(NULL);
        }
    }
}

void CubismClippingManager_GXM::CalcClippedDrawTotalBounds(CubismModel& model, CubismClippingContext* clippingContext)
{
    // 被クリッピングマスク（マスクされる描画オブジェクト）の全体の矩形
    csmFloat32 clippedDrawTotalMinX = FLT_MAX, clippedDrawTotalMinY = FLT_MAX;
    csmFloat32 clippedDrawTotalMaxX = FLT_MIN, clippedDrawTotalMaxY = FLT_MIN;

    // このマスクが実際に必要か判定する
    // このクリッピングを利用する「描画オブジェクト」がひとつでも使用可能であればマスクを生成する必要がある

    const csmInt32 clippedDrawCount = clippingContext->_clippedDrawableIndexList->GetSize();
    for (csmInt32 clippedDrawableIndex = 0; clippedDrawableIndex < clippedDrawCount; clippedDrawableIndex++)
    {
        // マスクを使用する描画オブジェクトの描画される矩形を求める
        const csmInt32 drawableIndex = (*clippingContext->_clippedDrawableIndexList)[clippedDrawableIndex];

        const csmInt32 drawableVertexCount = model.GetDrawableVertexCount(drawableIndex);
        csmFloat32* drawableVertexes = const_cast<csmFloat32*>(model.GetDrawableVertices(drawableIndex));

        csmFloat32 minX = FLT_MAX, minY = FLT_MAX;
        csmFloat32 maxX = FLT_MIN, maxY = FLT_MIN;

        csmInt32 loop = drawableVertexCount * Constant::VertexStep;
        for (csmInt32 pi = Constant::VertexOffset; pi < loop; pi += Constant::VertexStep)
        {
            csmFloat32 x = drawableVertexes[pi];
            csmFloat32 y = drawableVertexes[pi + 1];
            if (x < minX) minX = x;
            if (x > maxX) maxX = x;
            if (y < minY) minY = y;
            if (y > maxY) maxY = y;
        }

        //
        if (minX == FLT_MAX) continue; //有効な点がひとつも取れなかったのでスキップする

        // 全体の矩形に反映
        if (minX < clippedDrawTotalMinX) clippedDrawTotalMinX = minX;
        if (minY < clippedDrawTotalMinY) clippedDrawTotalMinY = minY;
        if (maxX > clippedDrawTotalMaxX) clippedDrawTotalMaxX = maxX;
        if (maxY > clippedDrawTotalMaxY) clippedDrawTotalMaxY = maxY;
    }
    if (clippedDrawTotalMinX == FLT_MAX)
    {
        clippingContext->_allClippedDrawRect->X = 0.0f;
        clippingContext->_allClippedDrawRect->Y = 0.0f;
        clippingContext->_allClippedDrawRect->Width = 0.0f;
        clippingContext->_allClippedDrawRect->Height = 0.0f;
        clippingContext->_isUsing = false;
    }
    else
    {
        clippingContext->_isUsing = true;
        csmFloat32 w = clippedDrawTotalMaxX - clippedDrawTotalMinX;
        csmFloat32 h = clippedDrawTotalMaxY - clippedDrawTotalMinY;
        clippingContext->_allClippedDrawRect->X = clippedDrawTotalMinX;
        clippingContext->_allClippedDrawRect->Y = clippedDrawTotalMinY;
        clippingContext->_allClippedDrawRect->Width = w;
        clippingContext->_allClippedDrawRect->Height = h;
    }
}

void CubismClippingManager_GXM::SetupLayoutBounds(csmInt32 usingClipCount) const
{
    if (usingClipCount <= 0)
    {// この場合は一つのマスクターゲットを毎回クリアして使用する
        for (csmUint32 index = 0; index < _clippingContextListForMask.GetSize(); index++)
        {
            CubismClippingContext* cc = _clippingContextListForMask[index];
            cc->_layoutChannelNo = 0; // どうせ毎回消すので固定で良い
            cc->_layoutBounds->X = 0.0f;
            cc->_layoutBounds->Y = 0.0f;
            cc->_layoutBounds->Width = 1.0f;
            cc->_layoutBounds->Height = 1.0f;
        }
        return;
    }

    // ひとつのRenderTextureを極力いっぱいに使ってマスクをレイアウトする
    // マスクグループの数が4以下ならRGBA各チャンネルに１つずつマスクを配置し、5以上6以下ならRGBAを2,2,1,1と配置する

    // RGBAを順番に使っていく。
    const csmInt32 div = usingClipCount / ColorChannelCount; //１チャンネルに配置する基本のマスク個数
    const csmInt32 mod = usingClipCount % ColorChannelCount; //余り、この番号のチャンネルまでに１つずつ配分する

    // RGBAそれぞれのチャンネルを用意していく(0:R , 1:G , 2:B, 3:A, )
    csmInt32 curClipIndex = 0; //順番に設定していく

    for (csmInt32 channelNo = 0; channelNo < ColorChannelCount; channelNo++)
    {
        // このチャンネルにレイアウトする数
        const csmInt32 layoutCount = div + (channelNo < mod ? 1 : 0);

        // 分割方法を決定する
        if (layoutCount == 0)
        {
            // 何もしない
        }
        else if (layoutCount == 1)
        {
            //全てをそのまま使う
            CubismClippingContext* cc = _clippingContextListForMask[curClipIndex++];
            cc->_layoutChannelNo = channelNo;
            cc->_layoutBounds->X = 0.0f;
            cc->_layoutBounds->Y = 0.0f;
            cc->_layoutBounds->Width = 1.0f;
            cc->_layoutBounds->Height = 1.0f;
        }
        else if (layoutCount == 2)
        {
            for (csmInt32 i = 0; i < layoutCount; i++)
            {
                const csmInt32 xpos = i % 2;

                CubismClippingContext* cc = _clippingContextListForMask[curClipIndex++];
                cc->_layoutChannelNo = channelNo;

                cc->_layoutBounds->X = xpos * 0.5f;
                cc->_layoutBounds->Y = 0.0f;
                cc->_layoutBounds->Width = 0.5f;
                cc->_layoutBounds->Height = 1.0f;
                //UVを2つに分解して使う
            }
        }
        else if (layoutCount <= 4)
        {
            //4分割して使う
            for (csmInt32 i = 0; i < layoutCount; i++)
            {
                const csmInt32 xpos = i % 2;
                const csmInt32 ypos = i / 2;

                CubismClippingContext* cc = _clippingContextListForMask[curClipIndex++];
                cc->_layoutChannelNo = channelNo;

                cc->_layoutBounds->X = xpos * 0.5f;
                cc->_layoutBounds->Y = ypos * 0.5f;
                cc->_layoutBounds->Width = 0.5f;
                cc->_layoutBounds->Height = 0.5f;
            }
        }
        else if (layoutCount <= 9)
        {
            //9分割して使う
            for (csmInt32 i = 0; i < layoutCount; i++)
            {
                const csmInt32 xpos = i % 3;
                const csmInt32 ypos = i / 3;

                CubismClippingContext* cc = _clippingContextListForMask[curClipIndex++];
                cc->_layoutChannelNo = channelNo;

                cc->_layoutBounds->X = xpos / 3.0f;
                cc->_layoutBounds->Y = ypos / 3.0f;
                cc->_layoutBounds->Width = 1.0f / 3.0f;
                cc->_layoutBounds->Height = 1.0f / 3.0f;
            }
        }
        else
        {
            CubismLogError("not supported mask count : %d", layoutCount);

            // 開発モードの場合は停止させる
            CSM_ASSERT(0);

            // 引き続き実行する場合、 SetupShaderProgramでオーバーアクセスが発生するので仕方なく適当に入れておく
            // もちろん描画結果はろくなことにならない
            for (csmInt32 i = 0; i < layoutCount; i++)
            {
                CubismClippingContext* cc = _clippingContextListForMask[curClipIndex++];
                cc->_layoutChannelNo = 0;
                cc->_layoutBounds->X = 0.0f;
                cc->_layoutBounds->Y = 0.0f;
                cc->_layoutBounds->Width = 1.0f;
                cc->_layoutBounds->Height = 1.0f;
            }
        }
    }
}

CubismRenderer::CubismTextureColor* CubismClippingManager_GXM::GetChannelFlagAsColor(csmInt32 channelNo)
{
    return _channelColors[channelNo];
}

csmVector<CubismClippingContext*>* CubismClippingManager_GXM::GetClippingContextListForDraw()
{
    return &_clippingContextListForDraw;
}

void CubismClippingManager_GXM::SetClippingMaskBufferSize(csmInt32 size)
{
    _clippingMaskBufferSize = size;
}

csmInt32 CubismClippingManager_GXM::GetClippingMaskBufferSize() const
{
    return _clippingMaskBufferSize;
}

/*********************************************************************************************************************
*                                      CubismClippingContext
********************************************************************************************************************/
CubismClippingContext::CubismClippingContext(CubismClippingManager_GXM* manager, const csmInt32* clippingDrawableIndices, csmInt32 clipCount)
{
    _owner = manager;

    // クリップしている（＝マスク用の）Drawableのインデックスリスト
    _clippingIdList = clippingDrawableIndices;

    // マスクの数
    _clippingIdCount = clipCount;

    _layoutChannelNo = 0;

    _allClippedDrawRect = CSM_NEW csmRectF();
    _layoutBounds = CSM_NEW csmRectF();

    _clippedDrawableIndexList = CSM_NEW csmVector<csmInt32>();
}

CubismClippingContext::~CubismClippingContext()
{
    if (_layoutBounds != NULL)
    {
        CSM_DELETE(_layoutBounds);
        _layoutBounds = NULL;
    }

    if (_allClippedDrawRect != NULL)
    {
        CSM_DELETE(_allClippedDrawRect);
        _allClippedDrawRect = NULL;
    }

    if (_clippedDrawableIndexList != NULL)
    {
        CSM_DELETE(_clippedDrawableIndexList);
        _clippedDrawableIndexList = NULL;
    }
}

void CubismClippingContext::AddClippedDrawable(csmInt32 drawableIndex)
{
    _clippedDrawableIndexList->PushBack(drawableIndex);
}

CubismClippingManager_GXM* CubismClippingContext::GetClippingManager()
{
    return _owner;
}

/*********************************************************************************************************************
*                                       CubismShader_GXM
********************************************************************************************************************/
namespace {
    const csmInt32 ShaderCount = 19; ///< シェーダの数 = マスク生成用 + (通常 + 加算 + 乗算) * (マスク無 + マスク有 + マスク有反転 + マスク無の乗算済アルファ対応版 + マスク有の乗算済アルファ対応版 + マスク有反転の乗算済アルファ対応版)
    CubismShader_GXM* s_instance;
    SceGxmShaderPatcher *s_shaderPatcher_GXM;
}

enum ShaderNames
{
    // SetupMask
    ShaderNames_SetupMask,

    //Normal
    ShaderNames_Normal,
    ShaderNames_NormalMasked,
    ShaderNames_NormalMaskedInverted,
    ShaderNames_NormalPremultipliedAlpha,
    ShaderNames_NormalMaskedPremultipliedAlpha,
    ShaderNames_NormalMaskedInvertedPremultipliedAlpha,

    //Add
    ShaderNames_Add,
    ShaderNames_AddMasked,
    ShaderNames_AddMaskedInverted,
    ShaderNames_AddPremultipliedAlpha,
    ShaderNames_AddMaskedPremultipliedAlpha,
    ShaderNames_AddMaskedPremultipliedAlphaInverted,

    //Mult
    ShaderNames_Mult,
    ShaderNames_MultMasked,
    ShaderNames_MultMaskedInverted,
    ShaderNames_MultPremultipliedAlpha,
    ShaderNames_MultMaskedPremultipliedAlpha,
    ShaderNames_MultMaskedPremultipliedAlphaInverted,
};

void CubismShader_GXM::ReleaseShaderProgram()
{
    for (csmUint32 i = 0; i < _shaderSets.GetSize(); i++)
    {
        if (_shaderSets[i]->VertexProgram)
        {
            sceGxmShaderPatcherReleaseVertexProgram(s_shaderPatcher_GXM, _shaderSets[i]->VertexProgram);
            sceGxmShaderPatcherUnregisterProgram(s_shaderPatcher_GXM, _shaderSets[i]->VertexProgramId);

            _shaderSets[i]->VertexProgram = NULL;
            _shaderSets[i]->VertexProgramId = 0;
        }

        if (_shaderSets[i]->FragmentProgram)
        {
            sceGxmShaderPatcherReleaseFragmentProgram(s_shaderPatcher_GXM, _shaderSets[i]->FragmentProgram);
            sceGxmShaderPatcherUnregisterProgram(s_shaderPatcher_GXM, _shaderSets[i]->FragmentProgramId);

            _shaderSets[i]->VertexProgram = NULL;
            _shaderSets[i]->VertexProgramId = 0;
        }

        CSM_DELETE(_shaderSets[i]);
    }
}

SCE_CDECL_BEGIN
    extern const SceGxmProgram _binary_VertShaderSrcSetupMask_gxp_start;
    extern const SceGxmProgram _binary_FragShaderSrcSetupMask_gxp_start;
    extern const SceGxmProgram _binary_VertShaderSrc_gxp_start;
    extern const SceGxmProgram _binary_VertShaderSrcMasked_gxp_start;
    extern const SceGxmProgram _binary_FragShaderSrc_gxp_start;
    extern const SceGxmProgram _binary_FragShaderSrcPremultipliedAlpha_gxp_start;
    extern const SceGxmProgram _binary_FragShaderSrcMask_gxp_start;
    extern const SceGxmProgram _binary_FragShaderSrcMaskInverted_gxp_start;
    extern const SceGxmProgram _binary_FragShaderSrcMaskPremultipliedAlpha_gxp_start;
    extern const SceGxmProgram _binary_FragShaderSrcMaskInvertedPremultipliedAlpha_gxp_start;

    static const SceGxmProgram* VertShaderSrcSetupMask = &_binary_VertShaderSrcSetupMask_gxp_start;
    static const SceGxmProgram* FragShaderSrcSetupMask = &_binary_FragShaderSrcSetupMask_gxp_start;
    static const SceGxmProgram* VertShaderSrc = &_binary_VertShaderSrc_gxp_start;
    static const SceGxmProgram* VertShaderSrcMasked = &_binary_VertShaderSrcMasked_gxp_start;
    static const SceGxmProgram* FragShaderSrc = &_binary_FragShaderSrc_gxp_start;
    static const SceGxmProgram* FragShaderSrcPremultipliedAlpha = &_binary_FragShaderSrcPremultipliedAlpha_gxp_start;
    static const SceGxmProgram* FragShaderSrcMask = &_binary_FragShaderSrcMask_gxp_start;
    static const SceGxmProgram* FragShaderSrcMaskInverted = &_binary_FragShaderSrcMaskInverted_gxp_start;
    static const SceGxmProgram* FragShaderSrcMaskPremultipliedAlpha = &_binary_FragShaderSrcMaskPremultipliedAlpha_gxp_start;
    static const SceGxmProgram* FragShaderSrcMaskInvertedPremultipliedAlpha = &_binary_FragShaderSrcMaskInvertedPremultipliedAlpha_gxp_start;
SCE_CDECL_END

CubismShader_GXM::CubismShader_GXM()
{ }

CubismShader_GXM::~CubismShader_GXM()
{
    ReleaseShaderProgram();
}

CubismShader_GXM* CubismShader_GXM::GetInstance()
{
    if (s_instance == NULL)
    {
        s_instance = CSM_NEW CubismShader_GXM();
    }
    return s_instance;
}

void CubismShader_GXM::DeleteInstance()
{
    if (s_instance)
    {
        CSM_DELETE_SELF(CubismShader_GXM, s_instance);
        s_instance = NULL;
        s_shaderPatcher_GXM = NULL;
    }
}

void CubismShader_GXM::GenerateShaders(CubismRenderer_GXM *renderer)
{
    for (csmInt32 i = 0; i < ShaderCount; i++)
    {
        _shaderSets.PushBack(CSM_NEW CubismShaderSet());
    }

    s_shaderPatcher_GXM = renderer->shaderPatcher_GXM;

    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, VertShaderSrcSetupMask, &_shaderSets[0]->VertexProgramId);
    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, FragShaderSrcSetupMask, &_shaderSets[0]->FragmentProgramId);

    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, VertShaderSrc, &_shaderSets[1]->VertexProgramId);
    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, FragShaderSrc, &_shaderSets[1]->FragmentProgramId);

    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, VertShaderSrcMasked, &_shaderSets[2]->VertexProgramId);
    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, FragShaderSrcMask, &_shaderSets[2]->FragmentProgramId);

    _shaderSets[3]->VertexProgramId = _shaderSets[2]->VertexProgramId;
    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, FragShaderSrcMaskInverted, &_shaderSets[3]->FragmentProgramId);

    _shaderSets[4]->VertexProgramId = _shaderSets[1]->VertexProgramId;
    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, FragShaderSrcPremultipliedAlpha, &_shaderSets[4]->FragmentProgramId);

    _shaderSets[5]->VertexProgramId = _shaderSets[2]->VertexProgramId;
    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, FragShaderSrcMaskPremultipliedAlpha, &_shaderSets[5]->FragmentProgramId);

    _shaderSets[6]->VertexProgramId = _shaderSets[2]->VertexProgramId;
    sceGxmShaderPatcherRegisterProgram(renderer->shaderPatcher_GXM, FragShaderSrcMaskInvertedPremultipliedAlpha, &_shaderSets[6]->FragmentProgramId);

    // 加算も通常と同じシェーダーを利用する
    _shaderSets[7]->VertexProgramId = _shaderSets[1]->VertexProgramId;
    _shaderSets[7]->FragmentProgramId = _shaderSets[1]->FragmentProgramId;
    _shaderSets[8]->VertexProgramId = _shaderSets[2]->VertexProgramId;
    _shaderSets[8]->FragmentProgramId = _shaderSets[2]->FragmentProgramId;
    _shaderSets[9]->VertexProgramId = _shaderSets[3]->VertexProgramId;
    _shaderSets[9]->FragmentProgramId = _shaderSets[3]->FragmentProgramId;
    _shaderSets[10]->VertexProgramId = _shaderSets[4]->VertexProgramId;
    _shaderSets[10]->FragmentProgramId = _shaderSets[4]->FragmentProgramId;
    _shaderSets[11]->VertexProgramId = _shaderSets[5]->VertexProgramId;
    _shaderSets[11]->FragmentProgramId = _shaderSets[5]->FragmentProgramId;
    _shaderSets[12]->VertexProgramId = _shaderSets[6]->VertexProgramId;
    _shaderSets[12]->FragmentProgramId = _shaderSets[6]->FragmentProgramId;

    // 乗算も通常と同じシェーダーを利用する
    _shaderSets[13]->VertexProgramId = _shaderSets[1]->VertexProgramId;
    _shaderSets[13]->FragmentProgramId = _shaderSets[1]->FragmentProgramId;
    _shaderSets[14]->VertexProgramId = _shaderSets[2]->VertexProgramId;
    _shaderSets[14]->FragmentProgramId = _shaderSets[2]->FragmentProgramId;
    _shaderSets[15]->VertexProgramId = _shaderSets[3]->VertexProgramId;
    _shaderSets[15]->FragmentProgramId = _shaderSets[3]->FragmentProgramId;
    _shaderSets[16]->VertexProgramId = _shaderSets[4]->VertexProgramId;
    _shaderSets[16]->FragmentProgramId = _shaderSets[4]->FragmentProgramId;
    _shaderSets[17]->VertexProgramId = _shaderSets[5]->VertexProgramId;
    _shaderSets[17]->FragmentProgramId = _shaderSets[5]->FragmentProgramId;
    _shaderSets[18]->VertexProgramId = _shaderSets[6]->VertexProgramId;
    _shaderSets[18]->FragmentProgramId = _shaderSets[6]->FragmentProgramId;


    // SetupMask
    _shaderSets[0]->AttributePositionLocation = sceGxmProgramFindParameterByName(VertShaderSrcSetupMask, "a_position");
    _shaderSets[0]->AttributeTexCoordLocation = sceGxmProgramFindParameterByName(VertShaderSrcSetupMask, "a_texCoord");
    _shaderSets[0]->SamplerTexture0Location = sceGxmProgramFindParameterByName(FragShaderSrcSetupMask, "s_texture0");
    _shaderSets[0]->UniformClipMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcSetupMask, "u_clipMatrix");
    _shaderSets[0]->UnifromChannelFlagLocation = sceGxmProgramFindParameterByName(FragShaderSrcSetupMask, "u_channelFlag");
    _shaderSets[0]->UniformBaseColorLocation = sceGxmProgramFindParameterByName(FragShaderSrcSetupMask, "u_baseColor");

    // 通常
    _shaderSets[1]->AttributePositionLocation = sceGxmProgramFindParameterByName(VertShaderSrc, "a_position");
    _shaderSets[1]->AttributeTexCoordLocation = sceGxmProgramFindParameterByName(VertShaderSrc, "a_texCoord");
    _shaderSets[1]->SamplerTexture0Location = sceGxmProgramFindParameterByName(FragShaderSrc, "s_texture0");
    _shaderSets[1]->UniformMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrc, "u_matrix");
    _shaderSets[1]->UniformBaseColorLocation = sceGxmProgramFindParameterByName(FragShaderSrc, "u_baseColor");

    // 通常（クリッピング）
    _shaderSets[2]->AttributePositionLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "a_position");
    _shaderSets[2]->AttributeTexCoordLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "a_texCoord");
    _shaderSets[2]->SamplerTexture0Location = sceGxmProgramFindParameterByName(FragShaderSrcMask, "s_texture0");
    _shaderSets[2]->SamplerTexture1Location = sceGxmProgramFindParameterByName(FragShaderSrcMask, "s_texture1");
    _shaderSets[2]->UniformMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "u_matrix");
    _shaderSets[2]->UniformClipMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "u_clipMatrix");
    _shaderSets[2]->UnifromChannelFlagLocation = sceGxmProgramFindParameterByName(FragShaderSrcMask, "u_channelFlag");
    _shaderSets[2]->UniformBaseColorLocation = sceGxmProgramFindParameterByName(FragShaderSrcMask, "u_baseColor");

    // 通常（クリッピング・反転）
    _shaderSets[3]->AttributePositionLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "a_position");
    _shaderSets[3]->AttributeTexCoordLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "a_texCoord");
    _shaderSets[3]->SamplerTexture0Location = sceGxmProgramFindParameterByName(FragShaderSrcMaskInverted, "s_texture0");
    _shaderSets[3]->SamplerTexture1Location = sceGxmProgramFindParameterByName(FragShaderSrcMaskInverted, "s_texture1");
    _shaderSets[3]->UniformMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "u_matrix");
    _shaderSets[3]->UniformClipMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "u_clipMatrix");
    _shaderSets[3]->UnifromChannelFlagLocation = sceGxmProgramFindParameterByName(FragShaderSrcMaskInverted, "u_channelFlag");
    _shaderSets[3]->UniformBaseColorLocation = sceGxmProgramFindParameterByName(FragShaderSrcMaskInverted, "u_baseColor");

    // 通常（PremultipliedAlpha）
    _shaderSets[4]->AttributePositionLocation = sceGxmProgramFindParameterByName(VertShaderSrc, "a_position");
    _shaderSets[4]->AttributeTexCoordLocation = sceGxmProgramFindParameterByName(VertShaderSrc, "a_texCoord");
    _shaderSets[4]->SamplerTexture0Location = sceGxmProgramFindParameterByName(FragShaderSrcPremultipliedAlpha, "s_texture0");
    _shaderSets[4]->UniformMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrc, "u_matrix");
    _shaderSets[4]->UniformBaseColorLocation = sceGxmProgramFindParameterByName(FragShaderSrcPremultipliedAlpha, "u_baseColor");

    // 通常（クリッピング、PremultipliedAlpha）
    _shaderSets[5]->AttributePositionLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "a_position");
    _shaderSets[5]->AttributeTexCoordLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "a_texCoord");
    _shaderSets[5]->SamplerTexture0Location = sceGxmProgramFindParameterByName(FragShaderSrcMaskPremultipliedAlpha, "s_texture0");
    _shaderSets[5]->SamplerTexture1Location = sceGxmProgramFindParameterByName(FragShaderSrcMaskPremultipliedAlpha, "s_texture1");
    _shaderSets[5]->UniformMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "u_matrix");
    _shaderSets[5]->UniformClipMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "u_clipMatrix");
    _shaderSets[5]->UnifromChannelFlagLocation = sceGxmProgramFindParameterByName(FragShaderSrcMaskPremultipliedAlpha, "u_channelFlag");
    _shaderSets[5]->UniformBaseColorLocation = sceGxmProgramFindParameterByName(FragShaderSrcMaskPremultipliedAlpha, "u_baseColor");

    // 通常（クリッピング・反転、PremultipliedAlpha）
    _shaderSets[6]->AttributePositionLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "a_position");
    _shaderSets[6]->AttributeTexCoordLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "a_texCoord");
    _shaderSets[6]->SamplerTexture0Location = sceGxmProgramFindParameterByName(FragShaderSrcMaskInvertedPremultipliedAlpha, "s_texture0");
    _shaderSets[6]->SamplerTexture1Location = sceGxmProgramFindParameterByName(FragShaderSrcMaskInvertedPremultipliedAlpha, "s_texture1");
    _shaderSets[6]->UniformMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "u_matrix");
    _shaderSets[6]->UniformClipMatrixLocation = sceGxmProgramFindParameterByName(VertShaderSrcMasked, "u_clipMatrix");
    _shaderSets[6]->UnifromChannelFlagLocation = sceGxmProgramFindParameterByName(FragShaderSrcMaskInvertedPremultipliedAlpha, "u_channelFlag");
    _shaderSets[6]->UniformBaseColorLocation = sceGxmProgramFindParameterByName(FragShaderSrcMaskInvertedPremultipliedAlpha, "u_baseColor");

    // 加算
    _shaderSets[7]->AttributePositionLocation = _shaderSets[1]->AttributePositionLocation;
    _shaderSets[7]->AttributeTexCoordLocation = _shaderSets[1]->AttributeTexCoordLocation;
    _shaderSets[7]->SamplerTexture0Location = _shaderSets[1]->SamplerTexture0Location;
    _shaderSets[7]->UniformMatrixLocation = _shaderSets[1]->UniformMatrixLocation;
    _shaderSets[7]->UniformBaseColorLocation = _shaderSets[1]->UniformBaseColorLocation;

    // 加算（クリッピング）
    _shaderSets[8]->AttributePositionLocation = _shaderSets[2]->AttributePositionLocation;
    _shaderSets[8]->AttributeTexCoordLocation = _shaderSets[2]->AttributeTexCoordLocation;
    _shaderSets[8]->SamplerTexture0Location = _shaderSets[2]->SamplerTexture0Location;
    _shaderSets[8]->SamplerTexture1Location = _shaderSets[2]->SamplerTexture1Location;
    _shaderSets[8]->UniformMatrixLocation = _shaderSets[2]->UniformMatrixLocation;
    _shaderSets[8]->UniformClipMatrixLocation = _shaderSets[2]->UniformClipMatrixLocation;
    _shaderSets[8]->UnifromChannelFlagLocation = _shaderSets[2]->UnifromChannelFlagLocation;
    _shaderSets[8]->UniformBaseColorLocation = _shaderSets[2]->UniformBaseColorLocation;

    // 加算（クリッピング・反転）
    _shaderSets[9]->AttributePositionLocation = _shaderSets[3]->AttributePositionLocation;
    _shaderSets[9]->AttributeTexCoordLocation = _shaderSets[3]->AttributeTexCoordLocation;
    _shaderSets[9]->SamplerTexture0Location = _shaderSets[3]->SamplerTexture0Location;
    _shaderSets[9]->SamplerTexture1Location = _shaderSets[3]->SamplerTexture1Location;
    _shaderSets[9]->UniformMatrixLocation = _shaderSets[3]->UniformMatrixLocation;
    _shaderSets[9]->UniformClipMatrixLocation = _shaderSets[3]->UniformClipMatrixLocation;
    _shaderSets[9]->UnifromChannelFlagLocation = _shaderSets[3]->UnifromChannelFlagLocation;
    _shaderSets[9]->UniformBaseColorLocation = _shaderSets[3]->UniformBaseColorLocation;

    // 加算（PremultipliedAlpha）
    _shaderSets[10]->AttributePositionLocation = _shaderSets[4]->AttributePositionLocation;
    _shaderSets[10]->AttributeTexCoordLocation = _shaderSets[4]->AttributeTexCoordLocation;
    _shaderSets[10]->SamplerTexture0Location = _shaderSets[4]->SamplerTexture0Location;
    _shaderSets[10]->UniformMatrixLocation = _shaderSets[4]->UniformMatrixLocation;
    _shaderSets[10]->UniformBaseColorLocation = _shaderSets[4]->UniformBaseColorLocation;

    // 加算（クリッピング、PremultipliedAlpha）
    _shaderSets[11]->AttributePositionLocation = _shaderSets[5]->AttributePositionLocation;
    _shaderSets[11]->AttributeTexCoordLocation = _shaderSets[5]->AttributeTexCoordLocation;
    _shaderSets[11]->SamplerTexture0Location = _shaderSets[5]->SamplerTexture0Location;
    _shaderSets[11]->SamplerTexture1Location = _shaderSets[5]->SamplerTexture1Location;
    _shaderSets[11]->UniformMatrixLocation = _shaderSets[5]->UniformMatrixLocation;
    _shaderSets[11]->UniformClipMatrixLocation = _shaderSets[5]->UniformClipMatrixLocation;
    _shaderSets[11]->UnifromChannelFlagLocation = _shaderSets[5]->UnifromChannelFlagLocation;
    _shaderSets[11]->UniformBaseColorLocation = _shaderSets[5]->UniformBaseColorLocation;

    // 加算（クリッピング・反転、PremultipliedAlpha）
    _shaderSets[12]->AttributePositionLocation = _shaderSets[6]->AttributePositionLocation;
    _shaderSets[12]->AttributeTexCoordLocation = _shaderSets[6]->AttributeTexCoordLocation;
    _shaderSets[12]->SamplerTexture0Location = _shaderSets[6]->SamplerTexture0Location;
    _shaderSets[12]->SamplerTexture1Location = _shaderSets[6]->SamplerTexture1Location;
    _shaderSets[12]->UniformMatrixLocation = _shaderSets[6]->UniformMatrixLocation;
    _shaderSets[12]->UniformClipMatrixLocation = _shaderSets[6]->UniformClipMatrixLocation;
    _shaderSets[12]->UnifromChannelFlagLocation = _shaderSets[6]->UnifromChannelFlagLocation;
    _shaderSets[12]->UniformBaseColorLocation = _shaderSets[6]->UniformBaseColorLocation;

    // 乗算
    _shaderSets[13]->AttributePositionLocation = _shaderSets[1]->AttributePositionLocation;
    _shaderSets[13]->AttributeTexCoordLocation = _shaderSets[1]->AttributeTexCoordLocation;
    _shaderSets[13]->SamplerTexture0Location = _shaderSets[1]->SamplerTexture0Location;
    _shaderSets[13]->UniformMatrixLocation = _shaderSets[1]->UniformMatrixLocation;
    _shaderSets[13]->UniformBaseColorLocation = _shaderSets[1]->UniformBaseColorLocation;

    // 乗算（クリッピング）
    _shaderSets[14]->AttributePositionLocation = _shaderSets[2]->AttributePositionLocation;
    _shaderSets[14]->AttributeTexCoordLocation = _shaderSets[2]->AttributeTexCoordLocation;
    _shaderSets[14]->SamplerTexture0Location = _shaderSets[2]->SamplerTexture0Location;
    _shaderSets[14]->SamplerTexture1Location = _shaderSets[2]->SamplerTexture1Location;
    _shaderSets[14]->UniformMatrixLocation = _shaderSets[2]->UniformMatrixLocation;
    _shaderSets[14]->UniformClipMatrixLocation = _shaderSets[2]->UniformClipMatrixLocation;
    _shaderSets[14]->UnifromChannelFlagLocation = _shaderSets[2]->UnifromChannelFlagLocation;
    _shaderSets[14]->UniformBaseColorLocation = _shaderSets[2]->UniformBaseColorLocation;

    // 乗算（クリッピング・反転）
    _shaderSets[15]->AttributePositionLocation = _shaderSets[3]->AttributePositionLocation;
    _shaderSets[15]->AttributeTexCoordLocation = _shaderSets[3]->AttributeTexCoordLocation;
    _shaderSets[15]->SamplerTexture0Location = _shaderSets[3]->SamplerTexture0Location;
    _shaderSets[15]->SamplerTexture1Location = _shaderSets[3]->SamplerTexture1Location;
    _shaderSets[15]->UniformMatrixLocation = _shaderSets[3]->UniformMatrixLocation;
    _shaderSets[15]->UniformClipMatrixLocation = _shaderSets[3]->UniformClipMatrixLocation;
    _shaderSets[15]->UnifromChannelFlagLocation = _shaderSets[3]->UnifromChannelFlagLocation;
    _shaderSets[15]->UniformBaseColorLocation = _shaderSets[3]->UniformBaseColorLocation;

    // 乗算（PremultipliedAlpha）
    _shaderSets[16]->AttributePositionLocation = _shaderSets[4]->AttributePositionLocation;
    _shaderSets[16]->AttributeTexCoordLocation = _shaderSets[4]->AttributeTexCoordLocation;
    _shaderSets[16]->SamplerTexture0Location = _shaderSets[4]->SamplerTexture0Location;
    _shaderSets[16]->UniformMatrixLocation = _shaderSets[4]->UniformMatrixLocation;
    _shaderSets[16]->UniformBaseColorLocation = _shaderSets[4]->UniformBaseColorLocation;

    // 乗算（クリッピング、PremultipliedAlpha）
    _shaderSets[17]->AttributePositionLocation = _shaderSets[5]->AttributePositionLocation;
    _shaderSets[17]->AttributeTexCoordLocation = _shaderSets[5]->AttributeTexCoordLocation;
    _shaderSets[17]->SamplerTexture0Location = _shaderSets[5]->SamplerTexture0Location;
    _shaderSets[17]->SamplerTexture1Location = _shaderSets[5]->SamplerTexture1Location;
    _shaderSets[17]->UniformMatrixLocation = _shaderSets[5]->UniformMatrixLocation;
    _shaderSets[17]->UniformClipMatrixLocation = _shaderSets[5]->UniformClipMatrixLocation;
    _shaderSets[17]->UnifromChannelFlagLocation = _shaderSets[5]->UnifromChannelFlagLocation;
    _shaderSets[17]->UniformBaseColorLocation = _shaderSets[5]->UniformBaseColorLocation;

    // 乗算（クリッピング・反転、PremultipliedAlpha）
    _shaderSets[18]->AttributePositionLocation = _shaderSets[6]->AttributePositionLocation;
    _shaderSets[18]->AttributeTexCoordLocation = _shaderSets[6]->AttributeTexCoordLocation;
    _shaderSets[18]->SamplerTexture0Location = _shaderSets[6]->SamplerTexture0Location;
    _shaderSets[18]->SamplerTexture1Location = _shaderSets[6]->SamplerTexture1Location;
    _shaderSets[18]->UniformMatrixLocation = _shaderSets[6]->UniformMatrixLocation;
    _shaderSets[18]->UniformClipMatrixLocation = _shaderSets[6]->UniformClipMatrixLocation;
    _shaderSets[18]->UnifromChannelFlagLocation = _shaderSets[6]->UnifromChannelFlagLocation;
    _shaderSets[18]->UniformBaseColorLocation = _shaderSets[6]->UniformBaseColorLocation;
}

void CubismShader_GXM::SetupShaderProgram(CubismRenderer_GXM* renderer, SceGxmTexture *gxmTexture
                                                , csmInt32 vertexCount, csmFloat32* vertexArray
                                                , csmFloat32* uvArray, csmFloat32 opacity
                                                , CubismRenderer::CubismBlendMode colorBlendMode
                                                , CubismRenderer::CubismTextureColor baseColor
                                                , csmBool isPremultipliedAlpha, CubismMatrix44 matrix4x4
                                                , csmBool invertedMask)
{
    if (_shaderSets.GetSize() == 0)
    {
        GenerateShaders(renderer);
    }

    // Blending
    SceGxmBlendInfo blendInfo;
    blendInfo.colorMask = SCE_GXM_COLOR_MASK_ALL;
    blendInfo.alphaFunc = SCE_GXM_BLEND_FUNC_ADD;
    blendInfo.colorFunc = SCE_GXM_BLEND_FUNC_ADD;

    if (renderer->GetClippingContextBufferForMask() != NULL) // マスク生成時
    {
        CubismShaderSet* shaderSet = _shaderSets[ShaderNames_SetupMask];

        blendInfo.colorSrc = SCE_GXM_BLEND_FACTOR_ZERO;
        blendInfo.colorDst = SCE_GXM_BLEND_FACTOR_ONE_MINUS_SRC_COLOR;
        blendInfo.alphaSrc = SCE_GXM_BLEND_FACTOR_ZERO;
        blendInfo.alphaDst = SCE_GXM_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;

        SceGxmVertexAttribute vertexAttributes[2];
        SceGxmVertexStream vertexStreams[2];
        vertexAttributes[0].streamIndex = 0;
        vertexAttributes[0].offset = 0;
        vertexAttributes[0].format = SCE_GXM_ATTRIBUTE_FORMAT_F32;
        vertexAttributes[0].componentCount = 2;
        vertexAttributes[0].regIndex = sceGxmProgramParameterGetResourceIndex(shaderSet->AttributePositionLocation);
        vertexStreams[0].stride = sizeof(csmFloat32) * 2;
        vertexStreams[0].indexSource = SCE_GXM_INDEX_SOURCE_INDEX_16BIT;

        vertexAttributes[1].streamIndex = 1;
        vertexAttributes[1].offset = 0;
        vertexAttributes[1].format = SCE_GXM_ATTRIBUTE_FORMAT_F32;
        vertexAttributes[1].componentCount = 2;
        vertexAttributes[1].regIndex = sceGxmProgramParameterGetResourceIndex(shaderSet->AttributeTexCoordLocation);
        vertexStreams[1].stride = sizeof(csmFloat32) * 2;
        vertexStreams[1].indexSource = SCE_GXM_INDEX_SOURCE_INDEX_16BIT;

        sceGxmShaderPatcherCreateVertexProgram(
            renderer->shaderPatcher_GXM,
            shaderSet->VertexProgramId,
            vertexAttributes,
            2,
            vertexStreams,
            2,
            &shaderSet->VertexProgram);

        sceGxmShaderPatcherCreateFragmentProgram(
            renderer->shaderPatcher_GXM,
            shaderSet->FragmentProgramId,
            SCE_GXM_OUTPUT_REGISTER_FORMAT_UCHAR4,
            renderer->msaa_GXM,
            &blendInfo,
            NULL,
            &shaderSet->FragmentProgram);

        sceGxmSetVertexProgram(renderer->immContext_GXM, shaderSet->VertexProgram);
        sceGxmSetFragmentProgram(renderer->immContext_GXM, shaderSet->FragmentProgram);

        //テクスチャ設定
        sceGxmSetFragmentTexture(renderer->immContext_GXM, 0, gxmTexture);

        // 頂点配列の設定
        uint32_t vertexArraySize = sizeof(csmFloat32) * vertexCount * 2;
        void* vertexArrayData = renderer->PoolMemalign(vertexArraySize, sizeof(csmFloat32) * 2);
        sceClibMemcpy(vertexArrayData, vertexArray, vertexArraySize);
        sceGxmSetVertexStream(renderer->immContext_GXM, 0, vertexArrayData);
        // テクスチャ頂点の設定
        uint32_t uvArraySize = vertexArraySize;
        void* uvArrayData = renderer->PoolMemalign(uvArraySize, sizeof(csmFloat32) * 2);
        sceClibMemcpy(uvArrayData, uvArray, uvArraySize);
        sceGxmSetVertexStream(renderer->immContext_GXM, 1, uvArrayData);

        // チャンネル
        const csmInt32 channelNo = renderer->GetClippingContextBufferForMask()->_layoutChannelNo;
        CubismRenderer::CubismTextureColor* colorChannel = renderer->GetClippingContextBufferForMask()->GetClippingManager()->GetChannelFlagAsColor(channelNo);

        float colorChannelTemp[4];
        void *fragmentDefaultBuffer;
        void *vertexDefaultBuffer;
        sceGxmReserveFragmentDefaultUniformBuffer(renderer->immContext_GXM, &fragmentDefaultBuffer);
        sceGxmReserveVertexDefaultUniformBuffer(renderer->immContext_GXM, &vertexDefaultBuffer);

        colorChannelTemp[0] = colorChannel->R;
        colorChannelTemp[1] = colorChannel->G;
        colorChannelTemp[2] = colorChannel->B;
        colorChannelTemp[3] = colorChannel->A;

        sceGxmSetUniformDataF(fragmentDefaultBuffer, shaderSet->UnifromChannelFlagLocation, 0, 4, colorChannelTemp);
        sceGxmSetUniformDataF(vertexDefaultBuffer, shaderSet->UniformClipMatrixLocation, 0, 16, renderer->GetClippingContextBufferForMask()->_matrixForMask.GetArray());

        csmRectF* rect = renderer->GetClippingContextBufferForMask()->_layoutBounds;

        float baseColorTemp[4];

        baseColorTemp[0] = rect->X * 2.0f - 1.0f;
        baseColorTemp[1] = rect->Y * 2.0f - 1.0f;
        baseColorTemp[2] = rect->GetRight() * 2.0f - 1.0f;
        baseColorTemp[3] = rect->GetBottom() * 2.0f - 1.0f;

        sceGxmSetUniformDataF(fragmentDefaultBuffer, shaderSet->UniformBaseColorLocation, 0, 4, baseColorTemp);
    }
    else // マスク生成以外の場合
    {
        const csmBool masked = renderer->GetClippingContextBufferForDraw() != NULL;  // この描画オブジェクトはマスク対象か
        const csmInt32 offset = (masked ? ( invertedMask ? 2 : 1 ) : 0) + (isPremultipliedAlpha ? 3 : 0);

        CubismShaderSet* shaderSet;
        switch (colorBlendMode)
        {
        case CubismRenderer::CubismBlendMode_Normal:
        default:
            shaderSet = _shaderSets[ShaderNames_Normal + offset];
            blendInfo.colorSrc = SCE_GXM_BLEND_FACTOR_ONE;
            blendInfo.colorDst = SCE_GXM_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
            blendInfo.alphaSrc = SCE_GXM_BLEND_FACTOR_ONE;
            blendInfo.alphaDst = SCE_GXM_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
            break;

        case CubismRenderer::CubismBlendMode_Additive:
            shaderSet = _shaderSets[ShaderNames_Add + offset];
            blendInfo.colorSrc = SCE_GXM_BLEND_FACTOR_ONE;
            blendInfo.colorDst = SCE_GXM_BLEND_FACTOR_ONE;
            blendInfo.alphaSrc = SCE_GXM_BLEND_FACTOR_ZERO;
            blendInfo.alphaDst = SCE_GXM_BLEND_FACTOR_ONE;
            break;

        case CubismRenderer::CubismBlendMode_Multiplicative:
            shaderSet = _shaderSets[ShaderNames_Mult + offset];
            blendInfo.colorSrc = SCE_GXM_BLEND_FACTOR_DST_COLOR;
            blendInfo.colorDst = SCE_GXM_BLEND_FACTOR_ONE_MINUS_SRC_ALPHA;
            blendInfo.alphaSrc = SCE_GXM_BLEND_FACTOR_ZERO;
            blendInfo.alphaDst = SCE_GXM_BLEND_FACTOR_ONE;
            break;
        }

        SceGxmVertexAttribute vertexAttributes[2];
        SceGxmVertexStream vertexStreams[2];
        vertexAttributes[0].streamIndex = 0;
        vertexAttributes[0].offset = 0;
        vertexAttributes[0].format = SCE_GXM_ATTRIBUTE_FORMAT_F32;
        vertexAttributes[0].componentCount = 2;
        vertexAttributes[0].regIndex = sceGxmProgramParameterGetResourceIndex(shaderSet->AttributePositionLocation);
        vertexStreams[0].stride = sizeof(csmFloat32) * 2;
        vertexStreams[0].indexSource = SCE_GXM_INDEX_SOURCE_INDEX_16BIT;

        vertexAttributes[1].streamIndex = 1;
        vertexAttributes[1].offset = 0;
        vertexAttributes[1].format = SCE_GXM_ATTRIBUTE_FORMAT_F32;
        vertexAttributes[1].componentCount = 2;
        vertexAttributes[1].regIndex = sceGxmProgramParameterGetResourceIndex(shaderSet->AttributeTexCoordLocation);
        vertexStreams[1].stride = sizeof(csmFloat32) * 2;
        vertexStreams[1].indexSource = SCE_GXM_INDEX_SOURCE_INDEX_16BIT;

        sceGxmShaderPatcherCreateVertexProgram(
            renderer->shaderPatcher_GXM,
            shaderSet->VertexProgramId,
            vertexAttributes,
            2,
            vertexStreams,
            2,
            &shaderSet->VertexProgram);

        sceGxmShaderPatcherCreateFragmentProgram(
            renderer->shaderPatcher_GXM,
            shaderSet->FragmentProgramId,
            SCE_GXM_OUTPUT_REGISTER_FORMAT_UCHAR4,
            renderer->msaa_GXM,
            &blendInfo,
            NULL,
            &shaderSet->FragmentProgram);

        sceGxmSetVertexProgram(renderer->immContext_GXM, shaderSet->VertexProgram);
        sceGxmSetFragmentProgram(renderer->immContext_GXM, shaderSet->FragmentProgram);

        // 頂点配列の設定
        uint32_t vertexArraySize = sizeof(csmFloat32) * vertexCount * 2;
        void* vertexArrayData = renderer->PoolMemalign(vertexArraySize, sizeof(csmFloat32) * 2);
        sceClibMemcpy(vertexArrayData, vertexArray, vertexArraySize);
        sceGxmSetVertexStream(renderer->immContext_GXM, 0, vertexArrayData);
        // テクスチャ頂点の設定
        uint32_t uvArraySize = vertexArraySize;
        void* uvArrayData = renderer->PoolMemalign(uvArraySize, sizeof(csmFloat32) * 2);
        sceClibMemcpy(uvArrayData, uvArray, uvArraySize);
        sceGxmSetVertexStream(renderer->immContext_GXM, 1, uvArrayData);

        void *fragmentDefaultBuffer;
        void *vertexDefaultBuffer;
        sceGxmReserveFragmentDefaultUniformBuffer(renderer->immContext_GXM, &fragmentDefaultBuffer);
        sceGxmReserveVertexDefaultUniformBuffer(renderer->immContext_GXM, &vertexDefaultBuffer);

        if (masked)
        {
            // frameBufferに書かれたテクスチャ
            SceGxmTexture *tex = renderer->_offscreenFrameBuffer.GetColorBuffer();

            sceGxmSetFragmentTexture(renderer->immContext_GXM, 1, tex);

            // View座標をClippingContextの座標に変換するための行列を設定
            sceGxmSetUniformDataF(vertexDefaultBuffer, shaderSet->UniformClipMatrixLocation, 0, 16, renderer->GetClippingContextBufferForDraw()->_matrixForDraw.GetArray());

            // 使用するカラーチャンネルを設定
            const csmInt32 channelNo = renderer->GetClippingContextBufferForDraw()->_layoutChannelNo;
            CubismRenderer::CubismTextureColor* colorChannel = renderer->GetClippingContextBufferForDraw()->GetClippingManager()->GetChannelFlagAsColor(channelNo);

            float colorChannelTemp[4];
            colorChannelTemp[0] = colorChannel->R;
            colorChannelTemp[1] = colorChannel->G;
            colorChannelTemp[2] = colorChannel->B;
            colorChannelTemp[3] = colorChannel->A;

            sceGxmSetUniformDataF(fragmentDefaultBuffer, shaderSet->UnifromChannelFlagLocation, 0, 4, colorChannelTemp);
        }

        //テクスチャ設定
        sceGxmSetFragmentTexture(renderer->immContext_GXM, 0, gxmTexture);

        //座標変換
        sceGxmSetUniformDataF(vertexDefaultBuffer, shaderSet->UniformMatrixLocation, 0, 16, matrix4x4.GetArray());

        float baseColorTemp[4];

        baseColorTemp[0] = baseColor.R;
        baseColorTemp[1] = baseColor.G;
        baseColorTemp[2] = baseColor.B;
        baseColorTemp[3] = baseColor.A;

        sceGxmSetUniformDataF(fragmentDefaultBuffer, shaderSet->UniformBaseColorLocation, 0, 4, baseColorTemp);
    }
}

/*********************************************************************************************************************
 *                                      CubismRenderer_GXM
 ********************************************************************************************************************/

CubismRenderer* CubismRenderer::Create(void *context)
{
    return CSM_NEW CubismRenderer_GXM((CubismRenderer_GXM::GXMContext *)context);
}

void CubismRenderer::StaticRelease()
{
    CubismRenderer_GXM::DoStaticRelease();
}

CubismRenderer_GXM::CubismRenderer_GXM(CubismRenderer_GXM::GXMContext *context) : _clippingManager(NULL)
                                                                                                  , _clippingContextBufferForMask(NULL)
                                                                                                  , _clippingContextBufferForDraw(NULL)
                                                                                                  , _poolIndex(0)
                                                                                                  , _poolSize(1 * 1024 * 1024)
{
    immContext_GXM = *context->immContext;
    renderTarget_GXM = *context->renderTarget;
    displayBufferData_GXM[0] = *context->displayBufferData[0];
    displayBufferData_GXM[1] = *context->displayBufferData[1];
    displayBufferSync_GXM[0] = *context->displayBufferSync[0];
    displayBufferSync_GXM[1] = *context->displayBufferSync[1];
    displayColorSurface_GXM[0] = *context->displayColorSurface[0];
    displayColorSurface_GXM[1] = *context->displayColorSurface[1];
    depthStencilSurface_GXM = *context->depthStencilSurface;
    shaderPatcher_GXM = *context->shaderPatcher;
    msaa_GXM = *context->msaa;
    displayWidth = *context->displayWidth;
    displayHeight = *context->displayHeight;

    sceGxmAllocDeviceMemLinux(
        SCE_GXM_DEVICE_HEAP_ID_USER_NC,
        SCE_GXM_MEMORY_ATTRIB_READ,
        _poolSize,
        4,
        &poolMem);

    // テクスチャ対応マップの容量を確保しておく.
    _textures.PrepareCapacity(32, true);
}

CubismRenderer_GXM::~CubismRenderer_GXM()
{
    sceGxmFreeDeviceMemLinux(poolMem);

    CSM_DELETE_SELF(CubismClippingManager_GXM, _clippingManager);

    if (_offscreenFrameBuffer.IsValid())
    {
        _offscreenFrameBuffer.DestroyOffscreenFrame();
    }
}

void CubismRenderer_GXM::DoStaticRelease()
{
    CubismShader_GXM::DeleteInstance();
}

void CubismRenderer_GXM::Initialize(CubismModel* model)
{
    if (model->IsUsingMasking())
    {
        _clippingManager = CSM_NEW CubismClippingManager_GXM();  //クリッピングマスク・バッファ前処理方式を初期化
        _clippingManager->Initialize(
            *model,
            model->GetDrawableCount(),
            model->GetDrawableMasks(),
            model->GetDrawableMaskCounts()
        );

        _offscreenFrameBuffer.CreateOffscreenFrame(_clippingManager->GetClippingMaskBufferSize(), _clippingManager->GetClippingMaskBufferSize(), &this->shaderPatcher_GXM, &this->msaa_GXM);
    }

    _sortedDrawableIndexList.Resize(model->GetDrawableCount(), 0);

    CubismRenderer::Initialize(model);  //親クラスの処理を呼ぶ
}

void CubismRenderer_GXM::SetDisplayIndex(csmInt32 displayIndex)
{
    _displayIndex = displayIndex;
}

void CubismRenderer_GXM::DoDrawModel()
{
    csmBool needsSceneReset = true;
    PoolReset();

    sceGxmSetFrontStencilFunc(immContext_GXM,
        SCE_GXM_STENCIL_FUNC_ALWAYS,
        SCE_GXM_STENCIL_OP_KEEP,
        SCE_GXM_STENCIL_OP_KEEP,
        SCE_GXM_STENCIL_OP_KEEP,
        0, 0);

    sceGxmSetFrontDepthFunc(immContext_GXM, SCE_GXM_DEPTH_FUNC_ALWAYS);
    sceGxmSetFrontDepthWriteEnable(immContext_GXM, SCE_GXM_DEPTH_WRITE_DISABLED);

    //------------ クリッピングマスク・バッファ前処理方式の場合 ------------
    if (_clippingManager != NULL)
    {
        // サイズが違う場合はここで作成しなおし
        if (_offscreenFrameBuffer.GetBufferWidth() != static_cast<csmUint32>(_clippingManager->GetClippingMaskBufferSize()) ||
            _offscreenFrameBuffer.GetBufferHeight() != static_cast<csmUint32>(_clippingManager->GetClippingMaskBufferSize()))
        {
            _offscreenFrameBuffer.DestroyOffscreenFrame();
            _offscreenFrameBuffer.CreateOffscreenFrame(
                static_cast<csmUint32>(_clippingManager->GetClippingMaskBufferSize()), static_cast<csmUint32>(_clippingManager->GetClippingMaskBufferSize()), &this->shaderPatcher_GXM, &this->msaa_GXM);
        }

        _clippingManager->SetupClippingContext(*GetModel(), this);
    }

    const csmInt32 drawableCount = GetModel()->GetDrawableCount();
    const csmInt32* renderOrder = GetModel()->GetDrawableRenderOrders();

    // インデックスを描画順でソート
    for (csmInt32 i = 0; i < drawableCount; ++i)
    {
        const csmInt32 order = renderOrder[i];
        _sortedDrawableIndexList[order] = i;
    }

    // 描画
    for (csmInt32 i = 0; i < drawableCount; ++i)
    {
        const csmInt32 drawableIndex = _sortedDrawableIndexList[i];

        // Drawableが表示状態でなければ処理をパスする
        if (!GetModel()->GetDrawableDynamicFlagIsVisible(drawableIndex))
        {
            continue;
        }

        // クリッピングマスク
        CubismClippingContext* clipContext = (_clippingManager != NULL)
            ? (*_clippingManager->GetClippingContextListForDraw())[drawableIndex]
            : NULL;

        if (clipContext != NULL && IsUsingHighPrecisionMask()) // マスクを書く必要がある
        {
            sceGxmEndScene(this->immContext_GXM, NULL, NULL);
            needsSceneReset = true;

            if(clipContext->_isUsing) // 書くことになっていた
            {
                _offscreenFrameBuffer.BeginDraw(&this->immContext_GXM);

                // マスクをクリアする
                // 1が無効（描かれない）領域、0が有効（描かれる）領域。（シェーダで Cd*Csで0に近い値をかけてマスクを作る。1をかけると何も起こらない）
                _offscreenFrameBuffer.Clear(1.0f, 1.0f, 1.0f, 1.0f, &this->immContext_GXM);
            }

            {
                const csmInt32 clipDrawCount = clipContext->_clippingIdCount;
                for (csmInt32 index = 0; index < clipDrawCount; index++)
                {
                    const csmInt32 clipDrawIndex = clipContext->_clippingIdList[index];

                    // 頂点情報が更新されておらず、信頼性がない場合は描画をパスする
                    if (!GetModel()->GetDrawableDynamicFlagVertexPositionsDidChange(clipDrawIndex))
                    {
                        continue;
                    }

                    IsCulling(GetModel()->GetDrawableCulling(clipDrawIndex) != 0);

                    // 今回専用の変換を適用して描く
                    // チャンネルも切り替える必要がある(A,R,G,B)
                    SetClippingContextBufferForMask(clipContext);
                    DrawMesh(
                        GetModel()->GetDrawableTextureIndices(clipDrawIndex),
                        GetModel()->GetDrawableVertexIndexCount(clipDrawIndex),
                        GetModel()->GetDrawableVertexCount(clipDrawIndex),
                        const_cast<csmUint16*>(GetModel()->GetDrawableVertexIndices(clipDrawIndex)),
                        const_cast<csmFloat32*>(GetModel()->GetDrawableVertices(clipDrawIndex)),
                        reinterpret_cast<csmFloat32*>(const_cast<Core::csmVector2*>(GetModel()->GetDrawableVertexUvs(clipDrawIndex))),
                        GetModel()->GetDrawableOpacity(clipDrawIndex),
                        CubismRenderer::CubismBlendMode_Normal,   //クリッピングは通常描画を強制
                        false // マスク生成時はクリッピングの反転使用は全く関係がない
                    );
                }
            }

            {
                // --- 後処理 ---
                _offscreenFrameBuffer.EndDraw(&this->immContext_GXM);
                SetClippingContextBufferForMask(NULL);
            }
        }

        // クリッピングマスクをセットする
        SetClippingContextBufferForDraw(clipContext);

        IsCulling(GetModel()->GetDrawableCulling(drawableIndex) != 0);

        if (needsSceneReset) {
            sceGxmBeginScene(
                this->immContext_GXM,
                0,
                this->renderTarget_GXM,
                NULL,
                NULL,
                this->displayBufferSync_GXM[_displayIndex],
                &this->displayColorSurface_GXM[_displayIndex],
                &this->depthStencilSurface_GXM);
            needsSceneReset = false;
        }

        csmInt32 indexNum = GetModel()->GetDrawableVertexIndexCount(drawableIndex);

        if (indexNum) {
            DrawMesh(
                GetModel()->GetDrawableTextureIndices(drawableIndex),
                indexNum,
                GetModel()->GetDrawableVertexCount(drawableIndex),
                const_cast<csmUint16*>(GetModel()->GetDrawableVertexIndices(drawableIndex)),
                const_cast<csmFloat32*>(GetModel()->GetDrawableVertices(drawableIndex)),
                reinterpret_cast<csmFloat32*>(const_cast<Core::csmVector2*>(GetModel()->GetDrawableVertexUvs(drawableIndex))),
                GetModel()->GetDrawableOpacity(drawableIndex),
                GetModel()->GetDrawableBlendMode(drawableIndex),
                GetModel()->GetDrawableInvertedMask(drawableIndex) // マスクを反転使用するか
            );
        }
    }

    sceGxmEndScene(this->immContext_GXM, NULL, NULL);

}

void CubismRenderer_GXM::DrawMesh(csmInt32 textureNo, csmInt32 indexCount, csmInt32 vertexCount
                                        , csmUint16* indexArray, csmFloat32* vertexArray, csmFloat32* uvArray
                                        , csmFloat32 opacity, CubismBlendMode colorBlendMode, csmBool invertedMask)
{

#ifndef CSM_DEBUG
    if (_textures[textureNo] == 0) return;    // モデルが参照するテクスチャがバインドされていない場合は描画をスキップする
#endif

    // 裏面描画の有効・無効
    if (IsCulling())
    {
        sceGxmSetCullMode(immContext_GXM, SCE_GXM_CULL_CW);
    }
    else
    {
        sceGxmSetCullMode(immContext_GXM, SCE_GXM_CULL_NONE);
    }

    CubismTextureColor modelColorRGBA = GetModelColor();

    if (GetClippingContextBufferForMask() == NULL) // マスク生成時以外
    {
        modelColorRGBA.A *= opacity;
        if (IsPremultipliedAlpha())
        {
            modelColorRGBA.R *= modelColorRGBA.A;
            modelColorRGBA.G *= modelColorRGBA.A;
            modelColorRGBA.B *= modelColorRGBA.A;
        }
    }

    SceGxmTexture *drawTexture;   // シェーダに渡すテクスチャID

    // テクスチャマップからバインド済みテクスチャIDを取得
    // バインドされていなければダミーのテクスチャIDをセットする
    if(_textures[textureNo] != 0)
    {
        drawTexture = _textures[textureNo];
    }
    else
    {
        //TODO: is this safe with GXM?
        drawTexture = NULL;
    }

    CubismShader_GXM::GetInstance()->SetupShaderProgram(
        this, drawTexture, vertexCount, vertexArray, uvArray
        , opacity, colorBlendMode, modelColorRGBA, IsPremultipliedAlpha()
        , GetMvpMatrix(), invertedMask
    );

    // ポリゴンメッシュを描画する
    uint32_t indexArraySize = sizeof(uint16_t) * indexCount;
    void* indexArrayData = this->PoolMemalign(indexArraySize, sizeof(uint16_t));
    sceClibMemcpy(indexArrayData, indexArray, indexArraySize);
    sceGxmDraw(this->immContext_GXM, SCE_GXM_PRIMITIVE_TRIANGLES, SCE_GXM_INDEX_FORMAT_U16, indexArrayData, indexCount);

    // 後処理
    SetClippingContextBufferForDraw(NULL);
    SetClippingContextBufferForMask(NULL);
}

void CubismRenderer_GXM::SaveProfile()
{
}

void CubismRenderer_GXM::RestoreProfile()
{
}

void CubismRenderer_GXM::BindTexture(csmUint32 modelTextureNo, SceGxmTexture *gxmTexture)
{
    _textures[modelTextureNo] = gxmTexture;
}

const csmMap<csmInt32, SceGxmTexture*>& CubismRenderer_GXM::GetBindedTextures() const
{
    return _textures;
}

void *CubismRenderer_GXM::PoolMalloc(uint32_t size)
{
    if ((_poolIndex + size) < _poolSize) {
        void *addr = (void *)((unsigned int)poolMem->mappedBase + _poolIndex);
        _poolIndex += size;
        return addr;
    }
    return NULL;
}

void *CubismRenderer_GXM::PoolMemalign(uint32_t size, uint32_t alignment)
{
    unsigned int new_index = (_poolIndex + alignment - 1) & ~(alignment - 1);
    if ((new_index + size) < _poolSize) {
        void *addr = (void *)((unsigned int)poolMem->mappedBase + new_index);
        _poolIndex = new_index + size;
        return addr;
    }
    return NULL;
}

void CubismRenderer_GXM::PoolReset()
{
    _poolIndex = 0;
}

void CubismRenderer_GXM::SetClippingMaskBufferSize(csmInt32 size)
{
    //FrameBufferのサイズを変更するためにインスタンスを破棄・再作成する
    CSM_DELETE_SELF(CubismClippingManager_GXM, _clippingManager);

    _clippingManager = CSM_NEW CubismClippingManager_GXM();

    _clippingManager->SetClippingMaskBufferSize(size);

    _clippingManager->Initialize(
        *GetModel(),
        GetModel()->GetDrawableCount(),
        GetModel()->GetDrawableMasks(),
        GetModel()->GetDrawableMaskCounts()
    );
}

csmInt32 CubismRenderer_GXM::GetClippingMaskBufferSize() const
{
    return _clippingManager->GetClippingMaskBufferSize();
}

void CubismRenderer_GXM::SetClippingContextBufferForMask(CubismClippingContext* clip)
{
    _clippingContextBufferForMask = clip;
}

CubismClippingContext* CubismRenderer_GXM::GetClippingContextBufferForMask() const
{
    return _clippingContextBufferForMask;
}

void CubismRenderer_GXM::SetClippingContextBufferForDraw(CubismClippingContext* clip)
{
    _clippingContextBufferForDraw = clip;
}

CubismClippingContext* CubismRenderer_GXM::GetClippingContextBufferForDraw() const
{
    return _clippingContextBufferForDraw;
}

}}}}

//------------ LIVE2D NAMESPACE ------------
